#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Output"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import OptimizeResult

from aragog.parser import Parameters
from aragog.solver import Evaluator

logger: logging.Logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from aragog.solver import Solver, State


class Output:
    """Stores inputs and outputs of the models."""

    def __init__(self, solver: Solver):
        self.solver: Solver = solver
        self.parameters: Parameters = solver.parameters
        self.solution: OptimizeResult = self.solver.solution
        self.evaluator: Evaluator = self.solver.evaluator
        self.state: State = self.solver.state

    @property
    def shape_basic(self) -> np.ndarray:
        """Shape of the basic data"""
        return np.array([self.solution.y.shape[0] + 1, self.solution.y.shape[1]])

    @property
    def shape_staggered(self) -> np.ndarray:
        """Shape of the staggered data"""
        return self.solution.y.shape

    @property
    def convective_heat_flux_basic(self) -> np.ndarray:
        """Convective heat flux"""
        return self.state.convective_heat_flux() * self.parameters.scalings.heat_flux

    @property
    def density_basic(self) -> np.ndarray:
        """Density"""
        return (
            self.evaluator.phase_basic.density()
            * np.ones(self.shape_basic)
            * self.parameters.scalings.density
        )

    @property
    def dTdr(self) -> np.ndarray:
        """dTdr"""
        return self.solver.state.dTdr() * self.parameters.scalings.temperature_gradient

    @property
    def dTdrs(self) -> np.ndarray:
        """dTdrs"""
        return (  # FIXME
            self.solver.evaluator.phase_basic.dTdrs()
            * self.parameters.scalings.temperature_gradient
        )

    @property
    def heat_capacity_basic(self) -> np.ndarray:
        """Heat capacity"""
        return (
            self.solver.evaluator.phase_basic.heat_capacity()
            * np.ones(self.shape_basic)
            * self.parameters.scalings.heat_capacity
        )

    @property
    def liquidus_K_staggered(self) -> np.ndarray:
        """Liquidus"""
        return self.evaluator.phases.mixed.liquidus() * self.parameters.scalings.temperature

    @property
    def melt_fraction_staggered(self) -> np.ndarray:
        """Melt fraction"""
        return self.solver.data.phase.melt_fraction(
            self.solution.y, self.data.mesh.staggered.eos.pressure
        ) * np.ones(self.shape_staggered)

    @property
    def radii_km_basic(self) -> np.ndarray:
        """Radii of the basic mesh in km"""
        return self.data.mesh.basic.radii * self.data.parameters.scalings.radius * 1.0e-3

    @property
    def pressure_GPa_basic(self) -> np.ndarray:
        """Pressure of the basic mesh in GPa"""
        return self.data.mesh.basic.eos.pressure * self.data.parameters.scalings.pressure * 1.0e-9

    @property
    def pressure_GPa_staggered(self) -> np.ndarray:
        """Pressure of the staggered mesh in GPa"""
        return (
            self.data.mesh.staggered.eos.pressure * self.data.parameters.scalings.pressure * 1.0e-9
        )

    @property
    def solidus_K_staggered(self) -> np.ndarray:
        """Solidus"""
        return (
            self.data.mixed.solidus(self.solution.y, self.data.mesh.staggered.eos.pressure)
            * self.data.parameters.scalings.temperature
        )

    @property
    def super_adiabatic_temperature_gradient_basic(self) -> np.ndarray:
        """Super adiabatic temperature gradient"""
        return (
            self.state.super_adiabatic_temperature_gradient
            * self.data.parameters.scalings.temperature_gradient
        )

    @property
    def temperature_K_basic(self) -> np.ndarray:
        """Temperature of the basic mesh in K"""
        return self.state.temperature_basic * self.data.parameters.scalings.temperature

    @property
    def temperature_K_staggered(self) -> np.ndarray:
        """Temperature of the staggered mesh in K"""
        return self.solver.temperature_staggered

    @property
    def thermal_expansivity_basic(self) -> np.ndarray:
        """Thermal expansivity"""
        return (
            self.solver.state.phase_basic.thermal_expansivity
            * np.ones(self.shape_basic)
            * self.data.parameters.scalings.thermal_expansivity
        )

    @property
    def log10_viscosity_basic(self) -> np.ndarray:
        """Viscosity of the basic mesh"""
        return np.log10(
            self.state.phase_basic.viscosity
            * self.data.parameters.scalings.viscosity
            * np.ones(self.shape_basic)
        )

    @property
    def times(self) -> np.ndarray:
        """Times in years"""
        return self.solution.t * self.data.parameters.scalings.time_years

    @property
    def time_range(self) -> float:
        return self.times[-1] - self.times[0]

    def plot(self, num_lines: int = 11) -> None:
        """Plots the solution with labelled lines according to time.

        Args:
            num_lines: Number of lines to plot. Defaults to 11.
        """
        assert self.solution is not None

        self.state.update(self.solution.y, self.solution.t)

        _, axs = plt.subplots(1, 10, sharey=True)

        # Ensure there are at least 2 lines to plot (first and last).
        num_lines = max(2, num_lines)

        # Calculate the time step based on the total number of lines.
        time_step: float = self.time_range / (num_lines - 1)

        # plot temperature
        try:
            axs[0].plot(self.liquidus_K_staggered, self.pressure_GPa_staggered, "k--")
            axs[0].plot(self.solidus_K_staggered, self.pressure_GPa_staggered, "k--")
        except AttributeError:
            pass

        # Plot the first line.
        def plot_times(ax, x: np.ndarray, y: np.ndarray) -> None:
            label_first: str = f"{self.times[0]:.2f}"
            ax.plot(x[:, 0], y, label=label_first)

            # Loop through the selected lines and plot each with a label.
            for i in range(1, num_lines - 1):
                desired_time: float = self.times[0] + i * time_step
                # Find the closest available time step.
                closest_time_index: np.intp = np.argmin(np.abs(self.times - desired_time))
                time: float = self.times[closest_time_index]
                label: str = f"{time:.2f}"  # Create a label based on the time.
                ax.plot(
                    x[:, closest_time_index],
                    y,
                    label=label,
                )

            # Plot the last line.
            times_end: float = self.times[-1]
            label_last: str = f"{times_end:.2f}"
            ax.plot(x[:, -1], y, label=label_last)

        plot_times(axs[0], self.temperature_K_basic, self.pressure_GPa_basic)
        axs[0].set_ylabel("Pressure (GPa)")
        axs[0].set_xlabel("Temperature (K)")
        axs[0].set_title("Temperature")

        plot_times(axs[1], self.melt_fraction_staggered, self.pressure_GPa_staggered)
        axs[1].set_xlabel("Melt fraction")
        axs[1].set_title("Melt fraction")

        plot_times(axs[2], self.log10_viscosity_basic, self.pressure_GPa_basic)
        axs[2].set_xlabel("Log10(viscosity)")
        axs[2].set_title("Log10(viscosity)")

        plot_times(axs[3], self.convective_heat_flux_basic, self.pressure_GPa_basic)
        axs[3].set_xlabel("Convective heat flux")
        axs[3].set_title("Convective heat flux")

        plot_times(
            axs[4],
            self.super_adiabatic_temperature_gradient_basic,
            self.pressure_GPa_basic,
        )
        axs[4].set_xlabel("Super adiabatic temperature gradient")
        axs[4].set_title("Super adiabatic temperature gradient")

        plot_times(
            axs[5],
            self.dTdr,
            self.pressure_GPa_basic,
        )
        axs[5].set_xlabel("dTdr")
        axs[5].set_title("dTdr")

        plot_times(
            axs[6],
            self.dTdrs,
            self.pressure_GPa_basic,
        )
        axs[6].set_xlabel("dTdrs")
        axs[6].set_title("dTdrs")

        plot_times(
            axs[7],
            self.density_basic,
            self.pressure_GPa_basic,
        )
        axs[7].set_xlabel("Density")
        axs[7].set_title("Density")

        plot_times(
            axs[8],
            self.heat_capacity_basic,
            self.pressure_GPa_basic,
        )
        axs[8].set_xlabel("Heat capacity")
        axs[8].set_title("Heat capacity")

        plot_times(
            axs[9],
            self.thermal_expansivity_basic,
            self.pressure_GPa_basic,
        )
        axs[9].set_xlabel("Thermal expansivity")
        axs[9].set_title("Thermal expansivity")

        # Shrink current axis by 20% to allow space for the legend.
        # box = ax.get_position()
        # ax.set_position((box.x0, box.y0, box.width * 0.8, box.height))

        # axs[0].grid(True)

        legend = axs[2].legend()  # (loc="center left", bbox_to_anchor=(1, 0.5))
        legend.set_title("Time (yr)")
        plt.gca().invert_yaxis()

        plt.show()



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Solver"""

from __future__ import annotations

import copy
import logging
from dataclasses import InitVar, dataclass, field
from pathlib import Path

import numpy as np
from scipy.integrate import solve_ivp
from scipy.optimize import OptimizeResult

from aragog.core import BoundaryConditions, InitialCondition
from aragog.interfaces import PhaseEvaluatorProtocol
from aragog.mesh import Mesh
from aragog.parser import Parameters, _EnergyParameters, _Radionuclide
from aragog.phase import PhaseEvaluatorCollection
from aragog.utilities import FloatOrArray

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class State:
    """Stores and updates the state at temperature and pressure.

    Args:
        parameters: Parameters
        evaluator: Evaluator

    Attributes:
        evaluator: Evaluator
        critical_reynolds_number: Critical Reynolds number
        gravitational_separation_flux: Gravitational separation flux at the basic nodes
        heating: Heat generation at the staggered nodes
        heat_flux: Heat flux at the basic nodes
        inviscid_regime: True if the flow is inviscid and otherwise False, at the basic nodes
        inviscid_velocity: Inviscid velocity at the basic nodes
        is_convective: True if the flow is convecting and otherwise False, at the basic nodes
        reynolds_number: Reynolds number at the basic nodes
        super_adiabatic_temperature_gradient: Super adiabatic temperature gradient at the basic nod
        temperature_basic: Temperature at the basic nodes
        temperature_staggered: Temperature at the staggered nodes
        bottom_temperature: Temperature at the bottom basic node
        top_temperature: Temperature at the top basic node
        viscous_regime: True if the flow is viscous and otherwise False, at the basic nodes
        viscous_velocity: Viscous velocity at the basic nodes
    """

    parameters: InitVar[Parameters]
    _evaluator: Evaluator
    _settings: _EnergyParameters = field(init=False)
    phase_basic: PhaseEvaluatorProtocol = field(init=False)
    phase_staggered: PhaseEvaluatorProtocol = field(init=False)
    _dTdr: np.ndarray = field(init=False)
    _eddy_diffusivity: np.ndarray = field(init=False)
    _heat_flux: np.ndarray = field(init=False)
    _heating: np.ndarray = field(init=False)
    _is_convective: np.ndarray = field(init=False)
    _reynolds_number: np.ndarray = field(init=False)
    _super_adiabatic_temperature_gradient: np.ndarray = field(init=False)
    _temperature_basic: np.ndarray = field(init=False)
    _temperature_staggered: np.ndarray = field(init=False)
    _viscous_velocity: np.ndarray = field(init=False)
    _inviscid_velocity: np.ndarray = field(init=False)

    def __post_init__(self, parameters: Parameters):
        self._settings = parameters.energy
        # Sets the pressure since this will not change during a model run. Must deepcopy first
        # because pressure is set as an attribute.
        self.phase_basic = copy.deepcopy(self._evaluator.phases.active)
        self.phase_basic.set_pressure(self._evaluator.mesh.basic.pressure)
        self.phase_staggered = copy.deepcopy(self._evaluator.phases.active)
        self.phase_staggered.set_pressure(self._evaluator.mesh.staggered.pressure)

    def capacitance_staggered(self) -> FloatOrArray:
        capacitance: FloatOrArray = (
            self.phase_staggered.density() * self.phase_staggered.heat_capacity()
        )

        return capacitance

    def conductive_heat_flux(self) -> np.ndarray:
        r"""Conductive heat flux:

        .. math::
            J_{cond} = -k \frac{\partial T}{\partial r}

        where :math:`k` is thermal conductivity, :math:`T` is temperature, and :math:`r` is radius.
        """
        conductive_heat_flux: np.ndarray = self.phase_basic.thermal_conductivity() * -self.dTdr()

        return conductive_heat_flux

    def convective_heat_flux(self) -> np.ndarray:
        r"""Convective heat flux:

        .. math::
            J_{conv} = -\rho c_p \kappa_h \left( \frac{\partial T}{\partial r} - \left( \frac{\partial T}{\partial r} \right)_S \right)

        where :math:`\rho` is density, :math:`c_p` is heat capacity at constant pressure,
        :math:`\kappa_h` is eddy diffusivity, :math:`T` is temperature, :math:`r` is radius, and
        :math:`S` is entropy.
        """
        convective_heat_flux: np.ndarray = (
            self.phase_basic.density()
            * self.phase_basic.heat_capacity()
            * self.eddy_diffusivity()
            * -self._super_adiabatic_temperature_gradient
        )

        return convective_heat_flux

    def radiogenic_heating(self, time: FloatOrArray) -> FloatOrArray:
        """Radiogenic heating

        Args:
            time: Time

        Returns:
            Radiogenic heating as a single column (in a 2-D array) if time is a float, otherwise a
                2-D array with each column associated with a single time in the time array.
        """
        radiogenic_heating_float: FloatOrArray = 0
        for radionuclide in self._evaluator.radionuclides:
            radiogenic_heating_float += radionuclide.get_heating(time)

        radiogenic_heating: FloatOrArray = radiogenic_heating_float * (
            self.phase_staggered.density() / self.capacitance_staggered()
        )

        return radiogenic_heating

    @property
    def critical_reynolds_number(self) -> float:
        """Critical Reynolds number from Abe (1993)"""
        return 9 / 8

    def dTdr(self) -> np.ndarray:
        return self._dTdr

    def eddy_diffusivity(self) -> np.ndarray:
        return self._eddy_diffusivity

    @property
    def gravitational_separation_flux(self) -> np.ndarray:
        """Gravitational separation"""
        raise NotImplementedError

    @property
    def heating(self) -> np.ndarray:
        """The total heating rate according to the heat sources specified in the configuration."""
        return self._heating

    @property
    def heat_flux(self) -> np.ndarray:
        """The total heat flux according to the fluxes specified in the configuration."""
        return self._heat_flux

    # TODO: Check this again
    @heat_flux.setter
    def heat_flux(self, value):
        """Setter for applying boundary conditions"""
        self._heat_flux = value

    @property
    def inviscid_regime(self) -> np.ndarray:
        return self._reynolds_number > self.critical_reynolds_number

    @property
    def inviscid_velocity(self) -> np.ndarray:
        return self._inviscid_velocity

    @property
    def is_convective(self) -> np.ndarray:
        return self._is_convective

    @property
    def mixing_flux(self) -> np.ndarray:
        """Mixing heat flux"""
        raise NotImplementedError

    @property
    def reynolds_number(self) -> np.ndarray:
        return self._reynolds_number

    @property
    def super_adiabatic_temperature_gradient(self) -> np.ndarray:
        return self._super_adiabatic_temperature_gradient

    @property
    def temperature_basic(self) -> np.ndarray:
        return self._temperature_basic

    @property
    def temperature_staggered(self) -> np.ndarray:
        return self._temperature_staggered

    @property
    def top_temperature(self) -> np.ndarray:
        return self._temperature_basic[-1, :]

    @property
    def bottom_temperature(self) -> np.ndarray:
        return self._temperature_basic[0, :]

    @property
    def viscous_regime(self) -> np.ndarray:
        return self._reynolds_number <= self.critical_reynolds_number

    @property
    def viscous_velocity(self) -> np.ndarray:
        return self._viscous_velocity

    def update(self, temperature: np.ndarray, time: FloatOrArray) -> None:
        """Updates the state.

        The evaluation order matters because we want to minimise the number of evaluations.

        Args:
            temperature: Temperature at the staggered nodes
            pressure: Pressure at the staggered nodes
            time: Time
        """
        logger.debug("Updating the state")

        logger.debug("Setting the temperature profile")
        self._temperature_staggered = temperature
        self._temperature_basic = self._evaluator.mesh.quantity_at_basic_nodes(temperature)
        logger.debug("temperature_basic = %s", self.temperature_basic)
        self._dTdr = self._evaluator.mesh.d_dr_at_basic_nodes(temperature)
        logger.debug("dTdr = %s", self.dTdr())
        self._evaluator.boundary_conditions.conform_temperature_boundary_conditions(
            temperature, self._temperature_basic, self.dTdr()
        )

        self.phase_staggered.set_temperature(temperature)
        self.phase_staggered.update()
        self.phase_basic.set_temperature(self._temperature_basic)
        self.phase_basic.update()

        self._super_adiabatic_temperature_gradient = self.dTdr() - self.phase_basic.dTdrs()
        self._is_convective = self._super_adiabatic_temperature_gradient < 0
        velocity_prefactor: np.ndarray = (
            self.phase_basic.gravitational_acceleration()
            * self.phase_basic.thermal_expansivity()
            * -self._super_adiabatic_temperature_gradient
        )
        # Viscous velocity
        self._viscous_velocity = (
            velocity_prefactor * self._evaluator.mesh.basic.mixing_length_cubed
        ) / (18 * self.phase_basic.kinematic_viscosity())
        self._viscous_velocity[~self.is_convective] = 0  # Must be super-adiabatic
        # Inviscid velocity
        self._inviscid_velocity = (
            velocity_prefactor * self._evaluator.mesh.basic.mixing_length_squared
        ) / 16
        self._inviscid_velocity[~self.is_convective] = 0  # Must be super-adiabatic
        self._inviscid_velocity[self._is_convective] = np.sqrt(
            self._inviscid_velocity[self._is_convective]
        )
        # Reynolds number
        self._reynolds_number = (
            self._viscous_velocity
            * self._evaluator.mesh.basic.mixing_length
            / self.phase_basic.kinematic_viscosity()
        )
        # Eddy diffusivity
        self._eddy_diffusivity = np.where(
            self.viscous_regime, self._viscous_velocity, self._inviscid_velocity
        )
        self._eddy_diffusivity *= self._evaluator.mesh.basic.mixing_length
        # Heat flux
        self._heat_flux = np.zeros_like(self.temperature_basic)
        if self._settings.conduction:
            self._heat_flux += self.conductive_heat_flux()
        if self._settings.convection:
            self._heat_flux += self.convective_heat_flux()
        if self._settings.gravitational_separation:
            self._heat_flux += self.gravitational_separation_flux
        if self._settings.mixing:
            self._heat_flux += self.mixing_flux
        # Heating
        self._heating = np.zeros_like(self.temperature_staggered)
        if self._settings.radionuclides:
            self._heating += self.radiogenic_heating(time)


@dataclass
class Evaluator:
    """Contains classes that evaluate quantities necessary to compute the interior evolution.

    Args:
        _parameters: Parameters

    Attributes:
        boundary_conditions: Boundary conditions
        initial_condition: Initial condition
        mesh: Mesh
        phases: Evaluators for all phases
        radionuclides: Radionuclides
    """

    _parameters: Parameters
    boundary_conditions: BoundaryConditions = field(init=False)
    initial_condition: InitialCondition = field(init=False)
    mesh: Mesh = field(init=False)
    phases: PhaseEvaluatorCollection = field(init=False)

    def __post_init__(self):
        self.mesh = Mesh(self._parameters)
        self.boundary_conditions = BoundaryConditions(self._parameters, self.mesh)
        self.initial_condition = InitialCondition(self._parameters, self.mesh)
        self.phases = PhaseEvaluatorCollection(self._parameters)

    @property
    def radionuclides(self) -> list[_Radionuclide]:
        return self._parameters.radionuclides


class Solver:
    """Solves the interior dynamics

    Args:
        filename: Filename of a file with configuration settings
        root: Root path to the flename

    Attributes:
        filename: Filename of a file with configuration settings
        root: Root path to the filename. Defaults to empty
        parameters: Parameters
        evaluator: Evaluator
        state: State
    """

    def __init__(self, filename: str | Path, root: str | Path = Path()):
        logger.info("Creating an Aragog model")
        self.filename = Path(filename)
        self.root = Path(root)
        self.parameters: Parameters
        self.evaluator: Evaluator
        self.state: State
        self._solution: OptimizeResult
        self.parse_configuration()

    def parse_configuration(self) -> None:
        """Parses a configuration file"""
        configuration_file: Path = self.root / self.filename
        logger.info("Parsing configuration file = %s", configuration_file)
        self.parameters = Parameters.from_file(configuration_file)

    def initialize(self) -> None:
        """Initializes the model."""
        logger.info("Initializing %s", self.__class__.__name__)
        self.evaluator = Evaluator(self.parameters)
        self.state = State(self.parameters, self.evaluator)

    @property
    def temperature_basic(self) -> np.ndarray:
        """Temperature of the basic mesh in K"""
        return self.evaluator.mesh.quantity_at_basic_nodes(self.temperature_staggered)

    @property
    def temperature_staggered(self) -> np.ndarray:
        """Temperature of the staggered mesh in K"""
        temperature: np.ndarray = self.solution.y * self.parameters.scalings.temperature

        return temperature

    @property
    def solution(self) -> OptimizeResult:
        """The solution."""
        return self._solution

    def dTdt(
        self,
        time: np.ndarray | float,
        temperature: np.ndarray,
    ) -> np.ndarray:
        """dT/dt at the staggered nodes

        Args:
            time: Time
            temperature: Temperature at the staggered nodes

        Returns:
            dT/dt at the staggered nodes
        """
        logger.debug("temperature passed into dTdt = %s", temperature)
        # logger.debug("temperature.shape = %s", temperature.shape)
        self.state.update(temperature, time)
        heat_flux: np.ndarray = self.state.heat_flux
        # logger.debug("heat_flux = %s", heat_flux)
        self.evaluator.boundary_conditions.apply(self.state)
        # logger.debug("heat_flux = %s", heat_flux)
        # logger.debug("mesh.basic.area.shape = %s", self.data.mesh.basic.area.shape)

        energy_flux: np.ndarray = heat_flux * self.evaluator.mesh.basic.area
        # logger.debug("energy_flux size = %s", energy_flux.shape)

        delta_energy_flux: np.ndarray = np.diff(energy_flux, axis=0)
        # logger.debug("delta_energy_flux size = %s", delta_energy_flux.shape)
        # logger.debug("capacitance = %s", self.state.phase_staggered.capacitance.shape)
        # FIXME: Update capacitance for mixed phase (enthalpy of fusion contribution)
        capacitance: np.ndarray = (
            self.state.capacitance_staggered() * self.evaluator.mesh.basic.volume
        )

        dTdt: np.ndarray = -delta_energy_flux / capacitance
        logger.debug("dTdt (fluxes only) = %s", dTdt)

        dTdt += self.state.heating
        logger.debug("dTdt (with internal heating) = %s", dTdt)

        return dTdt

    def solve(self) -> None:
        """Solves the system of ODEs to determine the interior temperature profile."""

        start_time: float = self.parameters.solver.start_time
        logger.debug("start_time = %f", start_time)
        end_time: float = self.parameters.solver.end_time
        logger.debug("end_time = %f", end_time)
        atol: float = self.parameters.solver.atol
        rtol: float = self.parameters.solver.rtol

        self._solution = solve_ivp(
            self.dTdt,
            (start_time, end_time),
            self.evaluator.initial_condition.temperature,
            method="BDF",
            vectorized=True,
            atol=atol,
            rtol=rtol,
        )

        logger.info(self.solution)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Mesh"""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from functools import cached_property

import numpy as np

from aragog.parser import Parameters, _MeshParameters
from aragog.utilities import FloatOrArray, is_monotonic_increasing

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class FixedMesh:
    """A fixed mesh

    Args:
        settings: Mesh parameters
        radii: Radii of the mesh
        outer_boundary: Outer boundary for computing depth below the surface
        inner_boundary: Inner boundary for computing height above the base

    Attributes:
        settings: Mesh parameters
        radii: Radii of the mesh
        outer_boundary: Outer boundary for computing depth below the surface. Defaults to None, in
            which case the outermost radius is used.
        inner_boundary: Inner boundary for computing height above the base. Defaults to None, in
            which case the innermost radius is used.
        area: Surface area
        delta_radii: Delta radii
        density: Density
        depth: Depth below the outer boundary
        height: Height above the inner boundary
        mixing_length: Mixing length
        mixing_length_squared: Mixing length squared
        mixing_length_cubed: Mixing length cubed
        number_of_nodes: Number of nodes
        pressure: Pressure
        pressure_gradient: Pressure gradient (dP/dr)
        volume: Volume of the spherical shells defined between neighbouring radii
        total_volume: Total volume
    """

    settings: _MeshParameters
    radii: np.ndarray
    outer_boundary: float | None = None
    inner_boundary: float | None = None
    _eos: AdamsWilliamsonEOS = field(init=False)

    def __post_init__(self):
        if not is_monotonic_increasing(self.radii):
            msg: str = "Mesh must be monotonically increasing"
            logger.error(msg)
            raise ValueError(msg)
        if self.outer_boundary is None:
            self.outer_boundary = np.max(self.radii)
        if self.inner_boundary is None:
            self.inner_boundary = np.min(self.radii)
        self._eos = AdamsWilliamsonEOS(
            self.settings, self.radii, self.outer_boundary, self.inner_boundary
        )

    @cached_property
    def area(self) -> np.ndarray:
        """Area"""
        return 4 * np.pi * np.square(self.radii)

    @cached_property
    def delta_radii(self) -> np.ndarray:
        return np.diff(self.radii, axis=0)

    @cached_property
    def density(self) -> np.ndarray:
        return self._eos.density

    @cached_property
    def depth(self) -> np.ndarray:
        return self.outer_boundary - self.radii

    @cached_property
    def height(self) -> np.ndarray:
        return self.radii - self.inner_boundary

    @cached_property
    def _mesh_cubed(self) -> np.ndarray:
        return np.power(self.radii, 3)

    @cached_property
    def mixing_length(self) -> np.ndarray:
        if self.settings.mixing_length_profile == "nearest_boundary":
            logger.debug("Set mixing length profile to nearest boundary")
            mixing_length = np.minimum(
                self.outer_boundary - self.radii, self.radii - self.inner_boundary
            )
        elif self.settings.mixing_length_profile == "constant":
            logger.debug("Set mixing length profile to constant")
            assert self.outer_boundary is not None
            assert self.inner_boundary is not None
            mixing_length = (
                np.ones_like(self.radii) * 0.25 * (self.outer_boundary - self.inner_boundary)
            )
        else:
            msg: str = f"Mixing length profile = {self.settings.mixing_length_profile} is unknown"
            raise ValueError(msg)

        return mixing_length

    @cached_property
    def mixing_length_cubed(self) -> np.ndarray:
        return np.power(self.mixing_length, 3)

    @cached_property
    def mixing_length_squared(self) -> np.ndarray:
        return np.square(self.mixing_length)

    @cached_property
    def number_of_nodes(self) -> int:
        return self.radii.size

    @cached_property
    def pressure(self) -> np.ndarray:
        return self._eos.pressure

    @cached_property
    def pressure_gradient(self) -> np.ndarray:
        return self._eos.pressure_gradient

    @cached_property
    def volume(self) -> np.ndarray:
        volume: np.ndarray = 4 / 3 * np.pi * (self._mesh_cubed[1:] - self._mesh_cubed[:-1])

        return volume

    @cached_property
    def total_volume(self) -> np.ndarray:
        return 4 / 3 * np.pi * (self._mesh_cubed[-1] - self._mesh_cubed[0])


class Mesh:
    """A staggered mesh.

    The basic mesh is used for the flux calculations and the staggered mesh is used for the volume
    calculations.

    Args:
        parameters: Parameters
    """

    def __init__(self, parameters: Parameters):
        self.settings: _MeshParameters = parameters.mesh
        basic_coordinates: np.ndarray = self.get_constant_spacing()
        self.basic: FixedMesh = FixedMesh(self.settings, basic_coordinates)
        staggered_coordinates: np.ndarray = self.basic.radii[:-1] + 0.5 * self.basic.delta_radii
        self.staggered: FixedMesh = FixedMesh(
            self.settings,
            staggered_coordinates,
            self.basic.outer_boundary,
            self.basic.inner_boundary,
        )
        self._d_dr_transform: np.ndarray = self._get_d_dr_transform_matrix()
        self._quantity_transform: np.ndarray = self._get_quantity_transform_matrix()

    def get_constant_spacing(self) -> np.ndarray:
        """Constant radius spacing across the mantle

        Returns:
            Radii with constant spacing as a column vector
        """
        radii: np.ndarray = np.linspace(
            self.settings.inner_radius, self.settings.outer_radius, self.settings.number_of_nodes
        )
        radii = np.atleast_2d(radii).T

        return radii

    def _get_d_dr_transform_matrix(self) -> np.ndarray:
        """Transform matrix for determining d/dr of a staggered quantity on the basic mesh.

        Returns:
            The transform matrix
        """
        transform: np.ndarray = np.zeros(
            (self.basic.number_of_nodes, self.staggered.number_of_nodes)
        )
        transform[1:-1, :-1] += np.diagflat(-1 / self.staggered.delta_radii)  # k=0 diagonal
        transform[1:-1:, 1:] += np.diagflat(1 / self.staggered.delta_radii)  # k=1 diagonal
        transform[0, :] = transform[1, :]  # Backward difference at outer radius
        transform[-1, :] = transform[-2, :]  # Forward difference at inner radius
        logger.debug("_d_dr_transform_matrix = %s", transform)

        return transform

    def d_dr_at_basic_nodes(self, staggered_quantity: np.ndarray) -> np.ndarray:
        """Determines d/dr at the basic nodes of a quantity defined at the staggered nodes.

        Args:
            staggered_quantity: A quantity defined at the staggered nodes.

        Returns:
            d/dr at the basic nodes
        """
        d_dr_at_basic_nodes: np.ndarray = self._d_dr_transform.dot(staggered_quantity)
        logger.debug("d_dr_at_basic_nodes = %s", d_dr_at_basic_nodes)

        return d_dr_at_basic_nodes

    # TODO: Compatibility with conforming boundary/initial conditions?
    def _get_quantity_transform_matrix(self) -> np.ndarray:
        """A transform matrix for mapping quantities on the staggered mesh to the basic mesh.

        Uses backward and forward differences at the inner and outer radius, respectively, to
        obtain the quantity values of the basic nodes at the innermost and outermost nodes. It may
        be subsequently necessary to conform these outer boundaries to applied boundary conditions.

        Returns:
            The transform matrix
        """
        transform: np.ndarray = np.zeros(
            (self.basic.number_of_nodes, self.staggered.number_of_nodes)
        )
        mesh_ratio: np.ndarray = self.basic.delta_radii[:-1] / self.staggered.delta_radii
        transform[1:-1, :-1] += np.diagflat(1 - 0.5 * mesh_ratio)  # k=0 diagonal
        transform[1:-1:, 1:] += np.diagflat(0.5 * mesh_ratio)  # k=1 diagonal
        # Backward difference at inner radius
        transform[0, :2] = np.array([1 + 0.5 * mesh_ratio[0], -0.5 * mesh_ratio[0]]).flatten()
        # Forward difference at outer radius
        mesh_ratio_outer: np.ndarray = self.basic.delta_radii[-1] / self.staggered.delta_radii[-1]
        transform[-1, -2:] = np.array(
            [-0.5 * mesh_ratio_outer, 1 + 0.5 * mesh_ratio_outer]
        ).flatten()
        logger.debug("_quantity_transform_matrix = %s", transform)

        return transform

    # TODO: Compatibility with conforming boundary/initial conditions?
    def quantity_at_basic_nodes(self, staggered_quantity: np.ndarray) -> np.ndarray:
        """Determines a quantity at the basic nodes that is defined at the staggered nodes.

        Uses backward and forward differences at the inner and outer radius, respectively, to
        obtain the quantity values of the basic nodes at the innermost and outermost nodes. It may
        be subsequently necessary to conform these outer boundaries to applied boundary conditions.

        Args:
            staggered_quantity: A quantity defined at the staggered nodes

        Returns:
            The quantity at the basic nodes
        """
        quantity_at_basic_nodes: np.ndarray = self._quantity_transform.dot(staggered_quantity)
        logger.debug("quantity_at_basic_nodes = %s", quantity_at_basic_nodes)

        return quantity_at_basic_nodes


class AdamsWilliamsonEOS:
    r"""Adams-Williamson equation of state (EOS).

    EOS due to adiabatic self-compression from the definition of the adiabatic bulk modulus:

    .. math::

        \left( \frac{{d\rho}}{{dP}} \right)_S = \frac{{\rho}}{{K_S}}

    where :math:`\rho` is density, :math:`K_S` the adiabatic bulk modulus, and :math:`S` is
    entropy.
    """

    def __init__(
        self,
        settings: _MeshParameters,
        radii: np.ndarray,
        outer_boundary: float,
        inner_boundary: float,
    ):
        self._settings: _MeshParameters = settings
        self._radii: np.ndarray = radii
        self._outer_boundary = outer_boundary
        self._inner_boundary = inner_boundary
        self._surface_density: float = self._settings.surface_density
        self._gravitational_acceleration: float = self._settings.gravitational_acceleration
        self._adiabatic_bulk_modulus: float = self._settings.adiabatic_bulk_modulus
        self._pressure = self.get_pressure_from_radii(radii)
        self._pressure_gradient = self.get_pressure_gradient(self.pressure)
        self._density = self.get_density(self.pressure)

    @property
    def density(self) -> np.ndarray:
        """Density"""
        return self._density

    @property
    def pressure(self) -> np.ndarray:
        """Pressure"""
        return self._pressure

    @property
    def pressure_gradient(self) -> np.ndarray:
        """Pressure gradient"""
        return self.pressure_gradient

    def get_density(self, pressure: FloatOrArray) -> np.ndarray:
        r"""Computes density from pressure:

        .. math::

            \rho(P) = \rho_s \exp(P/K_S)

        where :math:`\rho` is density, :math:`P` is pressure, :math:`\rho_s` is surface density,
        and :math:`K_S` is adiabatic bulk modulus.

        Args:
            pressure: Pressure

        Returns:
            Density
        """
        density: np.ndarray = self._surface_density * np.exp(
            pressure / self._adiabatic_bulk_modulus
        )

        return density

    def get_density_from_radii(self, radii: FloatOrArray) -> FloatOrArray:
        r"""Computes density from radii:

        .. math::

            \rho(r) = \frac{\rho_s K_S}{K_S + \rho_s g (r-r_s)}

        where :math:`\rho` is density, :math:`r` is radius, :math:`\rho_s` is surface density,
        :math:`K_S` is adiabatic bulk modulus, and :math:`r_s` is surface radius.

        Args:
            radii: Radii

        Returns
            Density
        """
        density: FloatOrArray = (self._surface_density * self._adiabatic_bulk_modulus) / (
            self._adiabatic_bulk_modulus
            + self._surface_density
            * self._gravitational_acceleration
            * (radii - self._outer_boundary)
        )

        return density

    def get_mass_element(self, radii: FloatOrArray) -> np.ndarray:
        r"""Computes the mass element:

        .. math::

            \frac{\delta m}{\delta r} = 4 \pi r^2 \rho

        where :math:`\delta m` is the mass element, :math:`r` is radius, and :math:`\rho` is
        density.

        Args:
            radii: Radii

        Returns:
            The mass element at radii
        """
        mass_element: np.ndarray = (
            4 * np.pi * np.square(radii) * self.get_density_from_radii(radii)
        )

        return mass_element

    def get_mass_within_radii(self, radii: FloatOrArray) -> np.ndarray:
        r"""Computes mass within radii:

        .. math::

            m(r) = \int 4 \pi r^2 \rho dr

        where :math:`m` is mass, :math:`r` is radius, and :math:`\rho` is density.

        The integral was evaluated using WolframAlpha.

        Args:
            radii: Radii

        Returns:
            Mass within radii
        """
        a: float = self._surface_density
        b: float = self._adiabatic_bulk_modulus
        c: float = self._gravitational_acceleration
        d: float = self._outer_boundary

        def mass_integral(radii_: FloatOrArray) -> np.ndarray:
            """Mass within radii including arbitrary constant of integration.

            Args:
                radii_: Radii

            Returns:
                Mass within radii
            """

            mass: np.ndarray = (
                4
                * np.pi
                * (
                    b
                    * (
                        a * c * radii_ * (a * c * (2 * d + radii_) - 2 * b)
                        + 2 * np.square(b - a * c * d) * np.log(a * c * (radii_ - d) + b)
                    )
                    / (2 * np.square(a) + np.power(c, 3))
                )
            )
            # + constant

            return mass

        mass: np.ndarray = mass_integral(radii) - mass_integral(self._inner_boundary)

        return mass

    def get_mass_within_shell(self, radii: np.ndarray) -> np.ndarray:
        """Computes the mass within spherical shells bounded by radii.

        Args:
            radii: Radii

        Returns:
            Mass within the bounded spherical shells
        """
        mass: np.ndarray = self.get_mass_within_radii(radii[1:]) - self.get_mass_within_radii(
            radii[:-1]
        )

        return mass

    def get_pressure_from_radii(self, radii: FloatOrArray) -> np.ndarray:
        r"""Computes pressure from radii:

        .. math::

            P(r) = -K_S \ln \left( 1 + \frac{\rho_s g (r-r_s)}{K_S} \right)

        where :math:`r` is radius, :math:`K_S` is adiabatic bulk modulus, :math:`P` is pressure,
        :math:`\rho_s` is surface density, :math:`g` is gravitational acceleration, and
        :math:`r_s` is surface radius.

        Args:
            radii: Radii

        Returns:
            Pressure
        """
        pressure: np.ndarray = -self._adiabatic_bulk_modulus * np.log(
            (
                self._adiabatic_bulk_modulus
                + self._surface_density
                * self._gravitational_acceleration
                * (radii - self._outer_boundary)
            )
            / self._adiabatic_bulk_modulus
        )

        return pressure

    def get_pressure_gradient(self, pressure: FloatOrArray) -> np.ndarray:
        r"""Computes the pressure gradient:

        .. math::

            \frac{dP}{dr} = -g \rho

        where :math:`\rho` is density, :math:`P` is pressure, and  :math:`g` is gravitational
        acceleration.

        Args:
            pressure: Pressure

        Returns:
            Pressure gradient
        """
        dPdr: np.ndarray = -self._gravitational_acceleration * self.get_density(pressure)

        return dPdr

    def get_radii_from_pressure(self, pressure: FloatOrArray) -> np.ndarray:
        r"""Computes radii from pressure:

        .. math::

            P(r) = \int \frac{dP}{dr} dr = \int -g \rho_s \exp(P/K_S) dr

        And apply the boundary condition :math:`P=0` at :math:`r=r_s` to get:

        .. math::

            r(P) = \frac{K_s \left( \exp(-P/K_S)-1 \right)}{\rho_s g} + r_s

        where :math:`r` is radius, :math:`K_S` is adiabatic bulk modulus, :math:`P` is pressure,
        :math:`\rho_s` is surface density, :math:`g` is gravitational acceleration, and
        :math:`r_s` is surface radius.

        Args:
            pressure: Pressure

        Returns:
            Radii
        """
        radii: np.ndarray = (
            self._adiabatic_bulk_modulus
            * (np.exp(-pressure / self._adiabatic_bulk_modulus) - 1)
            / (self._surface_density * self._gravitational_acceleration)
            + self._outer_boundary
        )

        return radii
